

#include "MatMult.h"
#include "SeqMult.h"
#include "SeqMultTrans.h"
#include "SeqMultBlock.h"
#include "SeqMultBlockTrans.h"
#include "ParMult.h"
#include "ParMultProc.h"
#include "ParMultProcTrans.h"
#include "ParMultProcBlock.h"
#include "ParMultProcBlockTrans.h"

#include "stdlib.h"
#include <iostream>
#include <sys/time.h>
#include <limits>

#include <math.h>

#define RAND_MAX 1;
#define THOUSAND 1000L;
#define MILLION 1000000L;

using namespace std;

/** 
 * Matrix Multiplication main class.
 * @author Joachim Staats.
 */

/** 
 * Performs a subtraction operation between two arrays sequentially.
 * @param C result matrix = A-B.
 * @param A left matrix.
 * @param B right matirx.
 */
void MatMult::seqdiff(double* C, int m, int n, double* A, double* B) 
{
    for (int i=0; i < m; i++) {
        for (int j=0; j < n; j++) {
            C[i*n+j] = A[i*n+j] - B[i*n+j];
        }
    }
}


/** 
 * Performs a transpose operation of the matrix.
 * Must be a square matrix.
 * @param A matrix.
 */
void MatMult::transposeInplace(double* A, int m, int n) 
{
    for (int i=0; i < m; i++) {
        for (int j=i+1; j < n; j++) {
            double aij = A[i*n+j];
            A[i*n+j] = A[j*n+i];
            A[j*n+i] = aij;
        }
    }
}


/** 
 * Performs a transpose operation of the matrix.
 * If A is n x m then B must be m x n.
 * @param A Array.
 * @param B Array transpose(A).
 */
void MatMult::transpose(double* B, double* A, int m, int n) 
{
    for (int i=0; i < m; i++) {
        for (int j=i+1; j < n; j++) {
            B[j*n+i] = A[i*n+j];
        }
    }
}


/**
 * Generates a matrix.
 * Creates a two dimensional array <CODE>A</CODE> with the dimension n, m.
 * The method <CODE>random</CODE> returns a 
 * random number within the intervall [0,1].
 * @param n dimension of matrix.
 * @param m dimension of matrix.
 */
double* MatMult::matgen(int m, int n) 
{
    double* A = new double[m*n];
    for (int i=0; i < m; i++) {
        for (int j=0; j < n; j++) {
            A[i*n+j] = rand();
        }
    }
    return A;  
}


/** 
 * Generates the null-matrix.
 * @param n dimension of matrix.
 * @param m dimension of matrix.
 */
double* MatMult::matgen0(int m, int n) 
{
    double* A = new double[m*n];
    for (int i=0; i < m; i++) {
        for (int j=0; j < n; j++) {
            A[i*n+j] = 0.0;
        }
    }
    return A;
}


/** 
 * Generates the unit-matrix.
 * @param n dimension of matrix.
 * @param m dimension of matrix.
 */
double* MatMult::matgen1(int m, int n) 
{
    double* A = new double[m*n];
    for (int i=0; i < m; i++) {
        for (int j=0; j < n; j++) {
            if (i == j) A[i*n+j] = 1.0; else A[i*n+j]= 0.0;
        }
    }
    return A;
}


/** 
 * Prints out the rows of matrix.
 * @param A Array. 
 */
void MatMult::matprint(double* A, int m, int n) 
{
    for (int i=0; i < m; i++) {
        for (int j=0; j < n; j++) {
            cout << A[i*n+j] << " ";
        }
        cout << endl;
    }
}


/** 
 * Checks if the matrix is approximatly zero.
 * Checks if the absolut value of the matrix's elements are less than
 * the production of the smallest positive value and 1000.
 * @param A Two dimensional array.
 */ 
bool MatMult::matcheck0(double* A, int m, int n) 
{
    double eps = numeric_limits<double>::min()*1000.0;
    for (int i=0; i < m; i++) {
        for (int j=0; j < n; j++) {
            if ( fabs(A[i*n+j]) > eps) { 
                cout << "(i,j)=(" << i << "," << j << ") = " << A[i*n+j];
                return false;
            }
        }
    }
    return true;
}

/** 
 * Makes a copy of matrix A.
 * @param A Two dimensional array.
 * @param B Two dimensional array.
 */ 
void MatMult::copy (double* A, double* B, int m, int n)
{
    for (int i=0; i < m; i++) {
        for (int j=0; j < n; j++) {
            B[i*n+j] = A[i*n+j];
        }
    }
}


/**
 * main
 */
int main() 
{
        int m = 800, // rows of A
            n = 800, // columns of B 
            p = 800; // columns of A = rows of B
        bool prnt = false;

/*        try { m = Integer.parseInt(args[0]); 
        } catch (Exception e) { }
        try { n = Integer.parseInt(args[1]); 
        } catch (Exception e) { }
        try { p = Integer.parseInt(args[2]); 
        } catch (Exception e) { }
        try { prnt = Boolean.valueOf(args[3]).booleanValue(); 
        } catch (Exception e) { }*/

        MatMult* x = new MatMult();

        SeqMult* seq    = new SeqMult();
/*        SeqMultBlock* seqb1  = new SeqMultBlock(15);*/
        SeqMultBlock* seqb2  = new SeqMultBlock(20);
/*        SeqMultBlock* seqb3  = new SeqMultBlock(25);*/
        SeqMultTrans* seqt   = new SeqMultTrans();
        SeqMultBlockTrans* seqbt  = new SeqMultBlockTrans(20);
        ParMult* par    = new ParMult();
        ParMultProc* parp   = new ParMultProc(8);
        ParMultProcBlock* parpb  = new ParMultProcBlock(8,20);
/*        ParMultProcBlock* parpb1 = new ParMultProcBlock(4,10);*/
        ParMultProcTrans* parpt  = new ParMultProcTrans(8);
        ParMultProcBlockTrans* parpbt = new ParMultProcBlockTrans(8,20);
        
        double* A = x->matgen(m,p);
        double* B = x->matgen(p,n);
        double* C  = x->matgen0(m,n);
        double* Cs = x->matgen0(m,n);
        double* D  = x->matgen0(m,n);
 
        cout << "Matrix size: " << m << " x " << p << " x " << n << endl;
        long flop = 1L * ((long)m) * (2L*(long)p) * ((long)n); 
        //                          mult + add
        long mflop = flop / 1000000L; 
        cout << "Operations: " << mflop << " MFLOP" << endl;
        if (prnt) {
           cout << "A = " << endl;
           x->matprint(A,m,p);
           cout << endl;
        }

        if (prnt) {
           cout << "B = " << endl;
           x->matprint(B,p,n);
           cout << endl;
        }

        struct timeval tv_start;
        struct timeval tv_end;

        long tm;         
        gettimeofday(&tv_start, NULL);
        seq->multiply(C,A,B,m,p,n);
        gettimeofday(&tv_end, NULL);
        tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
        tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
        cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;
        
        C = x->matgen0(m,n);
        gettimeofday(&tv_start, NULL);
        seqb2->multiply(C,A,B,m,p,n);
        gettimeofday(&tv_end, NULL);
        tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
        tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
        cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;

        if ( n == p ) {
           C = x->matgen0(m,n);
           gettimeofday(&tv_start, NULL);
           x->transposeInplace(B,p,n);
           seqt->multiply(C,A,B,m,p,n);
           gettimeofday(&tv_end, NULL);
           tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
           tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
           cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;

           x->transposeInplace(B,n,p); // required to count in next case

           gettimeofday(&tv_start, NULL);
           x->transposeInplace(B,p,n);
           seqbt->multiply(C,A,B,m,p,n);
           gettimeofday(&tv_end, NULL);
           tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
           tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
           cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;
        }

        x->copy(C,Cs,m,n);

        C = x->matgen0(m,n);
        gettimeofday(&tv_start, NULL);
        par->multiply(C,A,B,m,p,n);
        gettimeofday(&tv_end, NULL);
        tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
        tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
        cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;

        gettimeofday(&tv_start, NULL);
        parp->multiply(C,A,B,m,p,n);
        gettimeofday(&tv_end, NULL);
        tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
        tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
        cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;

        gettimeofday(&tv_start, NULL);
        parpb->multiply(C,A,B,m,p,n);
        gettimeofday(&tv_end, NULL);
        tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
        tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
        cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;

        if ( n == p ) {
           gettimeofday(&tv_start, NULL);
           x->transposeInplace(B,p,n);
           parpt->multiply(C,A,B,m,p,n);
           gettimeofday(&tv_end, NULL);
           tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
           tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
           cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;
           x->transposeInplace(B,p,n); // required to count in next case

           gettimeofday(&tv_start, NULL);
           x->transposeInplace(B,p,n);
           parpbt->multiply(C,A,B,m,p,n);
           gettimeofday(&tv_end, NULL);
           tm = ( tv_end.tv_sec  - tv_start.tv_sec )*THOUSAND;
           tm += ( tv_end.tv_usec - tv_start.tv_usec)/THOUSAND; 
           cout << tm << " ms, " << ((mflop*1000L/tm)) << " MFLOPS" << endl;
        }

        if (prnt) {
           cout << "Cs = " << endl;
           x->matprint(Cs,m,n);
           cout << endl;
        }

        if (prnt) {
           cout << "C = " << endl;
           x->matprint(C,m,n);
           cout << endl;
        }

        x->seqdiff(D,m,n,Cs,C); 
        if (prnt) {
           cout << "D = " << endl;
           x->matprint(D,m,n);
           cout << endl;
        }
        cout << "D is zero = " << x->matcheck0(D,m,n) << endl;
}
