
import java.util.Set;

import java.io.IOException;
import java.net.BindException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.ServerSocket;

import java.nio.channels.Selector;
import java.nio.channels.SelectionKey;
import java.nio.channels.ServerSocketChannel;
//import java.nio.channels.SocketChannel;


/**
 * ChannelFactoryNioBuffer.
 * Channel factory using java.nio and no Thread.
 * Deliveres ObjectChannels based on nio ByteBuffers.
 *
 * Usage: ChannelFactoryNioBuffer(port).
 * @author Heinz Kredel.
 */
public class ChannelFactoryNioBuffer extends Thread {
  
/**
 * default port of socket. 
 */
  public final static int DEFAULT_PORT = 4711;

/**
 * wait time 500ms. 
 */
  public final static int WAIT_TIME = 500;

/**
 * port of socket. 
 */
  private final int port;

/**
 * Capacity for ByteBuffer.
 */
  final int capacity; 

/**
 * local selector.
 */
  private Selector selector = null;

/**
 * local server key.
 */
  private SelectionKey serverkey = null;

/**
 * local server socket channel.
 */
  private ServerSocketChannel server = null;


/**
 * Constructs a ChannelFactory using Nio and ByteBuffer. 
 * @param p port.
 * @param cap capacity of ByteBuffer.
 */
  public ChannelFactoryNioBuffer(int p,int cap) {
    if ( p <= 0 ) { 
       port = DEFAULT_PORT; 
    } else { 
       port = p; 
    }
    if ( cap <= 0 ) { 
       capacity = ObjectBufferChannel.DEFAULT_CAPACITY; 
    } else { 
       capacity = cap; 
    }
    try {
        selector = Selector.open();

        server = ServerSocketChannel.open();
        server.configureBlocking(false);
        serverkey = server.register(selector,SelectionKey.OP_ACCEPT);
        server.socket().bind( new InetSocketAddress(port) );

        System.out.println("server started on port "+port);
    } catch (BindException e) { 
        System.out.println(e);
        serverkey.cancel();
        //server.close();
    } catch (IOException e) { 
        e.printStackTrace(); 
    }
  }


/**
 * Constructs a ChannelFactory using Nio and ByteBuffer.
 * On the DEFAULT_PORT with DEFAULT_CAPACITY.
 */
  public ChannelFactoryNioBuffer() {
      this(DEFAULT_PORT,ObjectBufferChannel.DEFAULT_CAPACITY); 
  }


/**
 * Constructs a ChannelFactory using Nio and ByteBuffer.
 * With DEFAULT_CAPACITY.
 * @param p port.
 */
  public ChannelFactoryNioBuffer(int p) {
      this(p,ObjectBufferChannel.DEFAULT_CAPACITY); 
  }


/**
 * Get a new socket channel from a server socket.
 */
  public ObjectBufferChannel getChannel() 
                             throws IOException {
    ObjectBufferChannel c = null;
    java.nio.channels.SocketChannel s;
    while (c == null) {
        try {
            System.out.println("waiting for connection");
            selector.select(); // blocking

            Set<SelectionKey> keys = selector.selectedKeys();
            for ( SelectionKey k: keys ) {
                if ( k == serverkey ) {
                   if ( k.isAcceptable() ) { // just in case
                      keys.remove(k);
                      s = server.accept();
                      s.configureBlocking(true);
                      c = new ObjectBufferChannel( s, capacity );
                      System.out.println("connection accepted");
                      break; // ignore other keys
                   }
                } else { // ignore client keys
                    selector.wakeup();
                }
            }
        } catch (IOException e) {
          System.out.println(e);
          try {   // wait for next client to connect
              Thread.currentThread().sleep( WAIT_TIME );
          } catch (InterruptedException ignored) { } 
        }
    }
    return c;
  }


/**
 * Get a new socket channel to a given host.
 * @param h hostname.
 * @param p port.
 */
  public ObjectBufferChannel getChannel(String h, int p) 
                                        throws IOException {
    if ( p <= 0 ) { 
       p = port; 
    } 
    SelectionKey clientkey = null;
    ObjectBufferChannel c = null;
    java.nio.channels.SocketChannel client;
    System.out.println("connecting to "+h);
    while (c == null) {
        try {
            client = java.nio.channels.SocketChannel.open();
            client.configureBlocking(false);
            clientkey = client.register(selector,SelectionKey.OP_CONNECT);
            client.connect( new InetSocketAddress(h,p) );

            selector.select(); // blocking

            Set<SelectionKey> keys = selector.selectedKeys();
            for ( SelectionKey k: keys ) {
                if ( k == clientkey ) {
                   if ( k.isConnectable() ) { // if server ready
                       keys.remove(k);
                       k.cancel();     // required to make blockable
                       client = (java.nio.channels.SocketChannel) k.channel();
                       client.finishConnect();
                       client.configureBlocking(true);
                       c = new ObjectBufferChannel( client, capacity );
                       break; // ignore other keys
                   }
                } else { // ignore server keys
                    selector.wakeup();
                }
            }
        } catch (IOException e) {
          System.out.println(e);
          System.out.println("Server on "+h+" not ready");
          try { // wait server ready
              Thread.currentThread().sleep( WAIT_TIME );
          } catch (InterruptedException ignored) { } 
        }
    }
    System.out.println("connected");
    return c;
  }


/**
 * Close the Channel Factory.
 */
  public void close() {
    try {
        serverkey.cancel();
        selector.close();
        server.close();
    } catch (IOException e) { 
        e.printStackTrace();
    }
  }

}

