

import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;

import java.nio.ByteBuffer;


/**
 * A wraper class to java.nio.ByteBuffer for handling objects.
 * The internal ByteBuffer is used for storing and retrieving objects.
 * The ByteBuffer may be direct or a heap buffer.
 * There are no bulk get / put methods since an object array can also 
 * be get / put as single object.
 * @author Heinz Kredel.
 */

public class ObjectBuffer {

    /*
     * The data structure.
     */
    private ByteBuffer bb;


    /**
     * Wrap an existing ByteBuffer as Object Buffer.
     */
    public ObjectBuffer(ByteBuffer bb) {
        this.bb = bb;
    }


    /**
     * Get and remove the next Object form the ByteBuffer.
     * If only a partial object remains in the buffer an 
     * BufferUnderFlowException is thrown.
     * @return the next object from the buffer.
     */
    public Object get() throws IOException, ClassNotFoundException {
        int pos;
        int rem;
        byte[] cont;
        if ( bb.hasArray() ) {
           cont = bb.array();
           pos = bb.position();
           rem = bb.remaining();
        } else {  // could be inefficient
           bb.mark();
           ByteBuffer bba = ByteBuffer.allocate( bb.remaining() );
           bba.clear();
           bba.put( bb );
           bba.rewind();
           cont = bba.array();
           pos = bba.position();
           rem = bba.remaining();
           bb.reset();
        }
        ByteArrayInputStream is  
            = new ByteArrayInputStream( cont, pos, rem );
        int read = is.available();
        ObjectInputStream ois = new ObjectInputStream( is );
        Object obj = ois.readObject();
        ois.close();
        read = read - is.available();
        //System.out.println("read = " + read);
        bb.position( bb.position() + read );
        //System.out.println("g-bb = " + bb);
        return obj;
    }


    /**
     * Put an Object to the ByteBuffer.
     * If the object does not fit into the buffer a 
     * BufferOverflowException is thrown.
     * @param obj the object to be put into the buffer.
     */
    public void put( Object obj ) throws IOException {
        ByteArrayOutputStream os  = new ByteArrayOutputStream();
        ObjectOutputStream oos = new ObjectOutputStream( os );
        oos.writeObject( obj );
        oos.close();
        //System.out.println("write = " + os.size());
        bb.put( os.toByteArray() );
        //System.out.println("p-bb = " + bb);
    }


    /**
     * to String
     */
    public String toString() {
        return "ObjectBuffer[" + bb.toString() + "]";
    }

}