
package algo;

import java.io.Serializable;

/**
 * A sequential algorithm for an euclidean 2d TSP Problem.
 * @author Heinz Kredel.
 */
public class SeqByteTSP implements TSPInf, Serializable {

    protected Graph graph;
    protected Point[] points;
    protected double best;   // beware of threads
    protected Path bestPath; // beware of threads
    protected long iter;
    protected long maxiter;


/**
 * @param p the cities.
 */
    public SeqByteTSP(Point[] p) {
        this.points = p;
        graph = new PlaneGraph(p);
        bestPath = PathByteArray.standardPath(graph); // standard path
        best = bestPath.cost();
        iter = 0; 
        maxiter = Long.MAX_VALUE;
    }


/**
 * @param p the cities.
 * @param g the corresponding graph.
 */
    public SeqByteTSP(Point[] p, Graph g) {
        this.points = p;
        graph = g;
        bestPath = PathByteArray.standardPath(graph); // standard path
        best = bestPath.cost();
        iter = 0; 
        maxiter = Long.MAX_VALUE;
    }


/**
 * @return number of iterations.
 */
    public long getIterations() {
        return iter;
    }


/**
 * @return maxiter.
 */
    public long getMaxIterations() {
        return maxiter;
    }


/**
 * @param m new maxIter.
 * @return old maxIter.
 */
    public long setMaxIterations(long m){
        long x = maxiter;
        maxiter = m;
        return x;
    }


/**
 * @return the actual best Path.
 */
    public Path actualBest() {
        return bestPath;
    }


/**
 * @param b a possible new best path.
 */
    public void setBest(Path b) {
        if ( b == null ) {
           return;
        }
        if ( b.cost() < best ) {
           bestPath = b;
           best = b.cost();
        }
    }


/**
 * @return getBest(Long.MAX_VALUE).
 */
    public Path getBest() {
        return getBest(Long.MAX_VALUE);
    }


/**
 * @param max maximal number of iterations.
 * @return best Path.
 */
    public Path getBest(long max) {
        iter = 0; 
        maxiter = max;
        bestPath = PathByteArray.standardPath(graph); // standard path
        best = bestPath.cost();
        if ( bestPath.length() <= 2 ) {
            return bestPath;
        }
        Path path = PathByteArray.onePath(graph); // one/empty path
        getBest(path);
        System.out.println("iterations = " + iter);
        return bestPath;
    }


/**
 * @param path a starting path.
 */
    public void getBest(Path path) {
        Path[] ps = path.nextPaths(); 
        iter += ps.length;
        if ( path.length()+1 == path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               //System.out.println("seq path = " + ps[i]);
               if ( ps[i].cost() < best ) {
                  best = ps[i].cost();
                  bestPath = ps[i];
                  System.out.println("seq best = " + bestPath);
               }
           }
           return;
        }
        if ( iter >= maxiter ) {
            return;
        }
        if ( path.length()+5 >= path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < best ) { // cut
                   //System.out.print("#");
                   getBestRec( ps[i].copyMax(), path.length()+1 ); // changes best and bestPath
                  //stack.add( ps[i] );
               }
           }
           return;
        }
        for ( int i = 0; i < ps.length; i++ ) { // branch
            //System.out.println("seq path part = " + ps[i]);
            if ( ps[i].cost() < best ) { // cut
               getBest( ps[i] ); // changes best and bestPath
            }
        }
        return;
    }


/**
 * @param path a starting path.
 * @param depth length of this path.
 */
    public void getBestRec(Path path, int depth) {
        for ( int k = 0; k < path.maxsize()-depth; k++ ) { // branch
            Path ps = path.nextPath(k); 
            iter++;
            if ( depth+1 == path.maxsize() ) {
               if ( ps.cost() < best ) {
                  best = ps.cost();
                  bestPath = (Path)ps.clone();
                  System.out.println("seq best = " + ps);
               }
            }
            if ( iter >= maxiter ) {
               return;
            }
            if ( ps.cost() < best ) { // cut
                getBestRec( ps, depth+1 ); // changes best and bestPath
                //stack.add( ps[i] );
            }
        }
        return;
    }

}